/*
 *  hitachi HD44780 4-bit parallel mode
 *  samsung KS0073  4-bit parallel or serial mode
 *  externally controlled lcd, async format "\1linea\2lineb\3etc"
 *
 *
 *  some timings for hitachi:
 *
 *  DBx latched by falling edge of E, setup time 195 nsec.
 *  RS (and hardwired RW) latched by E rising edge, setup time 140 nsec
 *  E minimum cycle time 1 usec, minumum pulse width 450 nsec
 *
 *  worst case delay during initialization
 *  and for Clear Display and Return Home, is 5 msec.
 *
 *  execution time _after_E_falls_
 *  is 120 usec worst case (with 160kHz osc ?)
 *  for commands other than Clear Display and Return Home.
 */
#include "standard.h"
#include "lcd.h"

word
	lcd_nrows,
	lcd_ncols,
	lcd_col,
	i2c_lcd_address = 0x4E;

STATIC enum { PAR, SER, EXT, I2C } interface;
STATIC enum { A00, A02, LM1110, KS0073 } chip;

STATIC word linekink;
STATIC byte rowpos[4], xlat[256];
STATIC void *cmd_switch, *data_switch;

#define cmd(x)  fncall((x), cmd_switch)
#define data(x) fncall((x), data_switch)

extern void
	lcd_cmd(), lcd_data(),
	ks0073_cmd(), ks0073_data(),
	i2c_lcd_cmd(), i2c_lcd_data(),
	SCI_putc();

void
lcd_chr(word c)
{
	if (++lcd_col == linekink)      /* linekink == 0 if not needed */
		cmd(0xC0);

	data(xlat[c]);
}

void
lcd_buf(byte *s, word cc)
{
	if (cc) {
		do {
			lcd_chr(*s++);
		} while (--cc);
	}
}

void
lcd_str(byte *s)
{
	lcd_buf(s, strlen(s));
}

/* 1...4 inclusive */

void
lcd_row(word which)
{
	if (which > lcd_nrows)
		which = lcd_nrows;

	lcd_col = 0;
	cmd(rowpos[which - 1]);
}

void
lcd_justify(int just)
{
	if (just < 0)
		just += lcd_ncols;

	if (just > 0)
		while (lcd_col < just)
			lcd_chr(' ');
}

void
lcd_finis()
{
	lcd_justify(lcd_ncols);
}

/* 9 pieces, but, depending on lcd, not all are used */

STATIC const byte
	cg_Aiti      [8] = { 0x0A, 0x00, 0x0E, 0x11, 0x1F, 0x11, 0x11, 0x00 },
	cg_Ake       [8] = { 0x04, 0x00, 0x0E, 0x11, 0x1F, 0x11, 0x11, 0x00 },
	cg_ake       [8] = { 0x04, 0x00, 0x0E, 0x01, 0x0F, 0x11, 0x0F, 0x00 },
	cg_padlock   [8] = { 0x00, 0x00, 0x00, 0x00, 0x06, 0x09, 0x0F, 0x0F },
	cg_flag      [8] = { 0x00, 0x08, 0x0C, 0x0E, 0x0F, 0x0B, 0x09, 0x08 },
	cg_backtick  [8] = { 0x08, 0x04, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00 },
	cg_backslash [8] = { 0x00, 0x10, 0x08, 0x04, 0x02, 0x01, 0x00, 0x00 },
	cg_tilde     [8] = { 0x00, 0x08, 0x15, 0x02, 0x00, 0x00, 0x00, 0x00 },
	cg_zero      [8];

STATIC word cg_ix;

/* create a glyph, return the character code */

static word
chargen(const byte *bits)
{
	int n = 8;

	cmd(0x40 + cg_ix * 8);   /* CG RAM address */

	do {
		data(*bits++);
	} while (--n);

	return (cg_ix++);
}

STATIC word prevssi;

void
lcd_rssi(word rssi)
{
	if (interface == EXT)
		return;

	rssi /= 50;       /* reasonable scale to [ 0 ... 8 ] */
	if (rssi > 8)
		rssi = 8;

	if (prevssi == rssi)
		return;

	/* XXX maybe do just one line per call ? might be nice */

	if (prevssi < rssi) {         /* pile grows or shrinks ? */

		cmd(0x40 + xlat[SIGMETER] * 8 + 8 - rssi);
		do {
			data(0xFF);
		} while (++prevssi < rssi);
	} else {
		cmd(0x40 + xlat[SIGMETER] * 8 + 8 - prevssi);
		do {
			data(0x00);
		} while (--prevssi > rssi);
	}
}

static void
charset()
{
	/* icons. leaves room for 5 custom characters */

	xlat[SIGMETER] = chargen(cg_zero);  prevssi = 0;
	xlat[FLAGUP]   = chargen(cg_flag);
	xlat[PADLOCK]  = chargen(cg_padlock);

	switch (chip) {

	default: /* Hitachi CG ROM Code: A00 */

		/* need all 5 remaining cg ram codes */

		xlat[0]       = 0xA5; /* invalid character, center dot */
		xlat[BULLSEP] = 0xA1; /* small square */
		xlat[BULLEOT] = 0xDB; /* large square */

		xlat[Aiti]  = chargen(cg_Aiti);
		xlat[aiti]  = 0x80 | 'a';
		xlat[Ake]   = chargen(cg_Ake);
		xlat[ake]   = chargen(cg_ake);
		xlat[Oljy]  =
		xlat[oljy]  = 0x80 | 'o'; /* lowercase close enough to uppercase */
		xlat[Uber]  =
		xlat[uber]  = 0x80 | 'u'; /* same here */

		xlat[degree] = 0x80 | '_';

		/* 7-bit ASCII almost correct */

		xlat['\\'] = chargen(cg_backslash);
		xlat['~']  = chargen(cg_tilde);
		break;

	case A02: /* Hitachi CG ROM Code: A02 */

		/* latin-1 as-is. very nice */

		xlat[0]       = 0xB7; /* center dot */
		xlat[BULLSEP] = 0x10; /* solid > */
		xlat[BULLEOT] = 0x11; /* solid < */
		break;

	case LM1110: /* 3rd Hitachi character set */

		xlat[0]       = 0xCD; /* center dot */
		xlat[BULLSEP] = 0xF6; /* solid > */
		xlat[BULLEOT] = 0xF7; /* solid < */

		xlat[Aiti]  = 0x8E;
		xlat[aiti]  = 0x84;
		xlat[Ake]   = 0x8F;
		xlat[ake]   = 0x86;
		xlat[Oljy]  = 0x99;
		xlat[oljy]  = 0x94;
		xlat[Uber]  = 0x9A;
		xlat[uber]  = 0x81;

		xlat[degree] = 0xB2;

		/* 7-bit part is ok as-is */
		break;

	case KS0073: /* Samsung KS0073 */

		xlat[0]       = 0xDD; /* center dot */
		xlat[BULLSEP] = 0x10; /* solid > */
		xlat[BULLEOT] = 0x11; /* sold < */

		xlat[Aiti]  = '[';
		xlat[aiti]  = '{';
		xlat[Ake]   = 0xAE;
		xlat[ake]   = 0xAF;
		xlat[Oljy]  = '\\';
		xlat[oljy]  = '|';
		xlat[Uber]  = '^';
		xlat[uber]  = '~';

		xlat[degree] = 0x80; /* super-0, close enough */

		/* fix 7-bit ASCII */

		xlat['$'] = 0xA2;
		xlat['@'] = 0xA0;

		xlat['['] = 0xFA;
		xlat['\\'] = 0xFB;
		xlat[']'] = 0xFC;
		xlat['^'] = 0x1D;
		xlat['_'] = 0xC4;
		xlat['`'] = chargen(cg_backtick);

		xlat['{'] = 0xFD;
		xlat['|'] = 0xFE;
		xlat['}'] = 0xFF;
		xlat['~'] = 0xCE;
		break;
	}

	/* XXX some translations for DOS and MAC */

	/* beware overwriting B0...FF */

	/*    MAC          DOS  */

	xlat[0x8C] = xlat[0x86] = xlat[ake];
	xlat[0x81] = xlat[0x8F] = xlat[Ake];

	xlat[0x8A] = xlat[0x84] = xlat[aiti];
	xlat[0x80] = xlat[0x8E] = xlat[Aiti];

	xlat[0x9A] = xlat[0x94] = xlat[oljy];
	xlat[0x85] = xlat[0x99] = xlat[Oljy];
}

void
lcd_init(int cfg_type, char *size)
{
	int i;

	for (i = 0; i < sizeof(xlat); ++i)
		xlat[i] = i;                   /* unity */

	lcd_nrows = atoi(size + 0);        /* size = "%1dx%d" */
	lcd_ncols = atoi(size + 2);

	cmd_switch  = lcd_cmd;
	data_switch = lcd_data;

	lcd_col   = 0;
	linekink  = 0;

	interface = PAR;
	chip      = A00;

	/* see tab_lcd_type[] */

	switch (cfg_type) {

	case 1:
		interface = EXT;
		break;

	case 3:
		chip = A02;
		break;

	case 4:
		chip = LM1110;
		break;

	case 6:
		interface = SER;
	case 5:
		chip = KS0073;
		break;
	}

	if (interface == PAR && i2c_probe(i2c_lcd_address))
		interface = I2C;

	switch (interface) {
	case SER:
		cmd_switch  = ks0073_cmd;
		data_switch = ks0073_data;
		break;
	case EXT:
		cmd_switch  = SCI_putc;
		data_switch = SCI_putc;
		break;
	case I2C:
		cmd_switch  = i2c_lcd_cmd;
		data_switch = i2c_lcd_data;
		break;
	}

	if (interface == EXT) {

		rowpos[0] = '\1'; /* in-band signalling */
		rowpos[1] = '\2';
		rowpos[2] = '\3';
		rowpos[3] = '\4';

		/* not fooling around with character sets or such */

		xlat[FLAGUP]  = '\20';
		xlat[PADLOCK] = '\21';
		xlat[BULLSEP] = '\22';
		xlat[BULLEOT] = '\23';

		xlat[0] = '.';
	} else {
		switch (chip) {

		default:
			rowpos[0] = 0x80 | 0x00;
			rowpos[1] = 0x80 | 0x40;
			rowpos[2] = 0x80 | 0x14;
			rowpos[3] = 0x80 | 0x54;
			break;

		case LM1110:
			rowpos[0] = 0x80;
			rowpos[1] = 0xC0;
			rowpos[2] = 0x90;
			rowpos[3] = 0xD0;
			break;

		case KS0073:
			rowpos[0] = 0x80 | 0x00;
			rowpos[1] = 0x80 | 0x20;
			rowpos[2] = 0x80 | 0x40;
			rowpos[3] = 0x80 | 0x60;
			break;
		}

		delay_ms(15); /* maybe needed in _re_boot */

		if (interface == PAR) {

			/* make sure interface is in 4-bit mode */

			i = 3;
			do {
				lcd_cmd8(0x38); /* Function Set, 8-bit interface */
				delay_ms(5);
			} while (--i);

			lcd_cmd8(0x28); /* Function Set, 4-bit interface */
			delay_ms(5);
		}
		if (chip == KS0073 && lcd_nrows == 4) {

			cmd(0x24); /* Function Set DL=0 N=0 RE=1 */
			cmd(0x09); /* Extended Function Set FW=0 BW=0 NW=1 */
			cmd(0x20); /* Function Set DL=0 N=0 RE=0 */
		}

		/* 1x16 is not really single line XXX correct in all cases ? */

		/* Function Set DL=0 N= F=0 */

		cmd(0x20 | ((lcd_nrows != 1 || lcd_ncols == 16) ? 8 : 0));

		cmd(0x06);     /* Entry Mode  I/D=1 S=0 increment, no shifting */
		cmd(0x0C);     /* Display ON, Blink OFF, Cursor OFF */
		cmd(0x02);     /* HOME */
		delay_ms(5);

		/* 1x16 (only) has kink, before 9th character */

		if (lcd_nrows == 1 && lcd_ncols == 16)
			linekink = 9;

		cg_ix = 0;
		charset();

		if (cg_ix > 8)
			doh("lcd cg"); /* too many custom characters needed */
	}
}

