;
;  one-wire data is MOSI and MISO, shorted together
;  (physically, or logically with buffering).
;  pullups required.
;
;  QSPAR.SCK does not exist. QSPI owns the pin when SPI is active,
;  when inactive, it is gpio according to DDRQS and PORTQS.
;  currently, only output pin in PORTQS is MOSI.
;  SCK is turned to input mode, it is still an OC output during spi'ing.
;  no problem as long as it is not read during ow routines.
;
;  or, maybe like this:
;  if ((SPCR1 & SPE) == 0)
;    portqs = PORTQS;
;  return (portqs & SCK);

	section .text
	align 1

SCK  equ 4  ; QS.2, is turned to input (mostly, OC output during onewire)
MOSI equ 2  ; QS.1, is onewire data and cannot be changed now, 
MISO equ 1  ; QS.0, is sacrificed to this.
 
QSMCR  equ 0FC00h  ; word
PORTQS equ 0FC15h  ; byte
PQSPAR equ 0FC16h  ; byte
DDRQS  equ 0FC17h  ; byte
SPCR0  equ 0FC18h  ; word
SPCR1  equ 0FC1Ah  ; word
SPCR2  equ 0FC1Ch  ; word
SPCR3  equ 0FC1Eh  ; byte
SPSR   equ 0FC1Fh  ; byte

RRAM   equ 0FD00h  ; word per entry, 16 entries
TRAM   equ 0FD20h  ; word per entry, 16 entries
CRAM   equ 0FD40h  ; byte per entry, 16 entries

; CRAM and some CRs are prepared in ow_reset

; SPI rate 0x22 about 4 usec per bit, about 65 usec per 16 bits
; SPI rate 0xFF about 30 usec per bit, 32 bits sent in reset, 972 usec

;  BITS IN ------------------------------------------

collect
	ldaa # 80h      ; carry after 8 rorred bits
L7
	ldab 0,y        ; MSByte is first 8 timeslots, 1F if "1" or 00 if "0"
	aiy  #+2
	addb # 0FCh     ; slice between 1FFF and 007F. carry if read-1
	rora            ; cumulate bits
	bcc  L7         ; if not yet 8 bits
	staa 0,x        ; bits get right-justified in last byte
	aix  #+1
	rts

	global _ow_bits_in ; (byte *bits, int nbits)

_ow_bits_in

	pshm k

	xgex            ; IX points to bit buffer in ram, XK assumed ok.
	xgdy            ; bit count in IY, 1 ... lots

	ldab # 0FFh
	tbek            ; EXT to registers
	tbyk            ; and Y too

	; fill transmit ram with 16 times write-1

	ldd  # 7FFFh    ; write-1 pattern
	tde             ; two patterns
	sted TRAM +  0
	sted TRAM +  4
	sted TRAM +  8
	sted TRAM + 12
	sted TRAM + 16
	sted TRAM + 20
	sted TRAM + 24
	sted TRAM + 28  ; doing 16 bits per spi burst, at most

	; send chunk of ones and look at receive ram

	xgey            ; bit count in E, 1 ... lots
L4
	ted             ; this many bits still to go, 1 ... lots
	cpd  # 16
	bls  L5
	ldd  # 16       ; doing by 16, at most
L5
	decb
	lbsr qspi_do_qp_and_q   ; D is index 0 ... 15 of last entry

	ldy  # RRAM
	cpe  # 8
	bls  L6         ; buf[1] must not be written
	bsr  collect    ; one byte
L6
	bsr  collect    ; another
	sube # 16
	bhi  L4         ; until no more bits

	pulm k
	rts

	global _ow_two_bits ; (void)

_ow_two_bits

	pshm k

	ldab # 0FFh
	tbek            ; EXT to registers

	ldd  # 7FFFh    ; write-1 pattern
	tde             ; two patterns
	sted TRAM       ; here

	ldd  # 1        ; entries 0 and 1
	bsr  qspi_do_qp_and_q

	ldd  # -1       ; error return is 4 or more

	lde  RRAM + 2   ; echo of second bit
	beq  L8         ; bus stuck low
	adde # 0FC00h   ; slice between 1FFF and 007F. carry if read-1
	rold

	lde  RRAM + 0   ; echo of first bit
	beq  L8
	adde # 0FC00h
	rold

	andd # 3        ; good return is 0 ... 3
L8
	pulm k
	rts


;  BITS OUT -----------------------------------------


	global _ow_bits_out ; (byte *bits, int nbits)

_ow_bits_out

	pshm k,z        ; using Z as scratch

	xgex            ; IX points to bit buffer in ram, XK assumed ok.
	xgde            ; bitcount in E (1...lots)

	ldab # 0FFh
	tbek            ; EXT to registers
	tbyk            ; and Y too
L60
	ldab 0,x        ; another 16 bits
	ldaa 1,x        ; order them to send LSbit first each byte
	aix  #+2        ; next 16
	bsr  out_some   ; bits in D, count of bits remaining in E
	tste
	bne  L60        ; while more bits to send

	pulm k,z
	rts

	; live regs X and E, D bits to write

out_some

	ldy  # TRAM - 2 ; start from entry 0, YK is set above
L61
	aiy  #+2
	ldz  # 0001h    ; assume "0"
	rord
	bcc  L62
	ldz  # 7FFFh    ; instead "1"
L62
	stz  0,y        ; Y left pointing to last, not past
	sube # 1
	beq  L63        ; no more bits
	cpy  # TRAM + 30
	bne  L61        ; doing at most 16 each time
L63
	xgdy            ; get pointer to last
	subd # TRAM     ; offset _to_ last
	lsrd            ; now index, 0 ... 15

qspi_do_qp_and_q    ; D (index) and CCR can change, only

	xgab            ; ENDQP is bits 0F00. NEWQP 000F and is now zero
	std   SPCR2     ; SPIFIE WREN WRTO ENDQP undefined NEWQP

	ldab # 22h      ; SPBR 0x22 is 60+ usec per 15 bits

	; fall thru

qspi_do_q           ; only D and CCR can change here

	ldaa # 0C3h     ; MSTR=1 WOMQ=1 BITS=0=16 CPOL=1 CPHA=1
	std  SPCR0      ; SPBR from B

	clrb
	stab  SPSR      ; SPIF cleared
	ldaa  # 80h     ; SPE=1
	std   SPCR1     ; GO !

	; XXX return now and wait SPIF at reset/out/in entrypoints
L64
	brclr SPSR, # 80h, L64
	bclr  SPSR, # 80h       ; wait SPIF and clear it when done

	rts             ; SPE clears automatically when done


; RESET --------------------------------------------

;  reset pulse min 480 usec max 960 usec
;  precense pulse 60-240 usec, 15-60 usec after end of reset pulse
;  min 480 usec idle from the end of reset pulse

	global _ow_reset

_ow_reset

	pshm  k
	ldab  # 0FFh
	tbek

	ldab # 0C3h     ; MSTR=1 WOMQ=1 BITS=0=16 CPOL=1 CPHA=1
	stab SPCR0      ; SPBR is not touched

	bclr  DDRQS,  # SCK  ; 0, input XXX just to use SCK as gpio input
	bclr  DDRQS,  # MISO
	bset  DDRQS,  # MOSI ; 1, data direction is output

	bset  PORTQS, # MOSI ; 1, High-Z it XXX fouls other pins ? only output.
	bset  PQSPAR, # MOSI ; 1, QSPI owns MOSI
	bset  PQSPAR, # MISO ; 1, and MISO

	ldab  # 40h     ; CONT BITSE=1 DT DSCK PCSx
	tba             ; two entries
	tde             ; four entries
	sted  CRAM +  0
	sted  CRAM +  4 ; eight
	sted  CRAM +  8
	sted  CRAM + 12 ; all sixteen

	clre            ; 0000
	ted
	comd            ; FFFF
	sted  TRAM + 0  ; two entries

	ldd   # 0100h   ; EDNQP=1 NEWQP=0
	std   SPCR2     ; SPIFIE WREN WRTO ENDQP undefined NEWQP
	ldab  # 0       ; LOOPQ=0
	stab  SPCR3     ; HMIE etc all 0

	ldab  # 0FFh    ; SPBR FF, reset pulse timing
	bsr   qspi_do_q

	ldd   RRAM + 2  ; presence pulse if any, here
	beq   L1        ; 0000, shorted to ground, likely
	comd
	beq   L1        ; FFFF, no devices, likely
	ldd   # 1       ; return D=1 TRUE if some present
L1
	pulm  k
	rts

